#include "system.h"
//#include <StrMini.h>
//#include <MMSystem.h>
//#include <KsMedia.h>
#include "DbgMsg.h"
#include "DbgFltr.h"
#include "margierr.h"
#include "Idents.h"

#include "MidasApi.h"
#include "Lsi64020.h"
#include "dvHwMain.h"
#include "Global.h"
#include "IdentCls.h"
#include "AuxData.h"

//30Sep98 Chip Doran
/*************************************************************************/
/** AuxData.c								**/
/** --	This module reads and sorts data from the LSI64020 Auxilliary	**/
/**	Data FIFO. Data from the AuxData contains such things are the	**/
/**	current picture type, GOP type, bitrates, etc.			**/
/** --	The AUXCMD_CHECK command needs to be called frequently,		**/
/**	typically during the timer tick in order to avoid loss of data.	**/
/*************************************************************************/

#define LAYER_NONE	0xFF	//Indicates no layers yet read.
#define LAYER_SEQ	0x00	//Sequence layer
#define LAYER_GOP	0x01	//Group Of Pictures
#define LAYER_PIC	0x02	//Picture
#define LAYER_PKT	0x04	//Packet
#define LAYER_EXT	0x07	//Extension

//Note that extension IDs are the bitstream ID plus LAYER_EXT.
#define LAYER_EXTSEQ	(LAYER_EXT+0x01)  //Sequence extension
#define LAYER_EXTSDISP	(LAYER_EXT+0x02)  //Sequence display extension
#define LAYER_EXTQTBL	(LAYER_EXT+0x03)  //Quant Matrix Extension
#define LAYER_EXTCOPYRT	(LAYER_EXT+0x04)  //Copyright Extension
#define LAYER_EXTPDISP	(LAYER_EXT+0x07)  //Picture Display Extension
#define LAYER_EXTPIC	(LAYER_EXT+0x08)  //Picture Coding Extension

/*************************************************************************/
/** Structure definitions.						**/
/** -	The layer structures come from the LSI64020 reference manual,	**/
/**	Chapter 8.							**/
/*************************************************************************/
#pragma pack(1)
struct LayerSeq_s {	//LAYER_SEQ. Table 8.1
  WORD Wid;		//horizontal_size_value (12bits)
  WORD Hgt;		//vertical_size_value (12bits)
  BYTE Aspect;		//aspect_ratio_information (4bits)
  BYTE FrameRate;	//frame_rate_code (4bits)
  WORD BitRateLo;	//Lower 16bits of bit_rate_value
  BYTE BitRateHi;	//Upper 2bits of bit_rate_value (18bits)
  WORD BufSize;		//vbv_buffer_size_value (10bits)
  BYTE Constrain;	//constrained_parameters_flag (1bit)
};

struct LayerGop_s {	//LAYER_GOP. Table 8.4
  DWORD TimeCode;	//time_code (25bits)
  BYTE Closed;		//closed_gop (1bit)
  BYTE Broken;		//broken_gop (1bit)
};

struct LayerPic_s {	//LAYER_PIC. Table 8.5
  WORD TimeRef;		//temporal_reference (10bits)
  BYTE PicType;		//picture_coding_type (3bits)
  WORD Delay;		//vbv_delay (16bits)
};

struct LayerPkt_s {	//LAYER_PKT
  BYTE Unk;		//I can't find this structure.
};

struct LayerExtSeq_s {	//LAYER_EXTSEQ. Table 8.2
  BYTE Profile;		//profile_and_level_indication (8bits)
  BYTE Progressive;	//progressive_sequence (1bit)
  BYTE ChromaFmt;	//chroma_format (2bits)
  BYTE WidExt;		//horizontal_size_extension (2bits)
  BYTE HgtExt;		//vertical_size_extension (2bits)
  WORD BitRateExt;	//bit_rate_extension (12bits)
  BYTE BufSizeExt;	//vbv_buffer_size_extension (8bits)
  BYTE LowDelay;	//low_delay (1bit)
  BYTE FrameRateExtN;	//frame_rate_extension_n (2bits)
  BYTE FrameRateExtD;	//frame_rate_extension_d (5bits)
};

struct LayerExtSeqDisp_s {  //LAYER_EXTSDISP. Table 8.3. **When Color==0**
  BYTE VidFmt;		//video_format (3bits)
  BYTE Color;		//color_description (1bit)
  WORD Wid;		//display_horizontal_size (14bits)
  WORD Hgt;		//display_vertical_size (14bits)
};

struct LayerExtSeqDispColor_s { //LAYER_EXTSDISP. Table 8.3. **When Color==1**
  BYTE VidFmt;		//video_format (3bits)
  BYTE Color;		//color_description (1bit)
  BYTE Primaries;	//color_primaries (8bits)
  BYTE Transfer;	//transfer_characteristics (8bits)
  BYTE Matrix;		//matrix_coefficients (8bits)
  WORD Wid;		//display_horizontal_size (14bits)
  WORD Hgt;		//display_vertical_size (14bits)
};

struct LayerExtCopyright_s {  //LAYER_EXTCOPYRT. Table 8.10
  BYTE Flag;		//copyright_flag (1bit)
  BYTE ID;		//copyright_identifier (8bits)
  BYTE IsCopy;		//original_or_copy (1bit)
  BYTE Rsvd;		//reserved (7bits)
  BYTE Number[9];	//copyright_number
};

//This table must be dealt with specially!!!! Its size depends on parameters from other layers!
struct LayerExtPicDisp_s {  //LAYER_EXTPDISP. Table 8.8
  WORD HorzOffsetLo;	//Lower 14bits of frame_center_horizontal_offset
  BYTE HorzOffsetHi;	//Upper 2bits of frame_center_horizontal_offset (16bits)
  WORD VertOffsetLo;	//Lower 12bits of frame_center_vertical_offset
  BYTE VertOffsetHi;	//Upper 4 bits (16bits)
};
#ifdef _NEC_
  #pragma pack(8)
#else
  #pragma pack()
#endif

//To start recording a layer, add it to the following list.
static struct LayersInUse_s {
  BYTE LayerID;
  BYTE Len;
  DWORD Timeout;	//If the layer is expected on a regular basis (ie Picture layer), otherwise 0. (msec)
} LayersInUse[]= {
  { LAYER_SEQ, sizeof(struct LayerSeq_s), 0 },
  { LAYER_GOP, sizeof(struct LayerGop_s), 0 },
  { LAYER_PIC, sizeof(struct LayerPic_s), 70 }	//Expected every 42ms. Over 84ms means we missed one. (Unless resuming from pause.)
};
#define LAYER_MAX (sizeof(LayersInUse)/sizeof(LayersInUse[0]))

typedef struct DataLayer_s {	//Common header for all layers.
  BYTE ID;		//Since not all sequential values are used, I use an ID lookup.
  BYTE Len;		//Length in bytes of complete data layer. 0 is legal for variable-length layers.
  BYTE CurByte;		//Number of bytes written to layer. Complete when CurByte>=Len.
  DWORD Timeout;	//Timeout value. May be 0.
  DWORD TimeoutCnt;	//Total number of timeouts seen in this layer.
  DWORD PcTime;		//PC time at which layer was started.
  union {		//Acutal data.
    struct LayerSeq_s Seq;
    struct LayerGop_s Gop;
    struct LayerPic_s Pic;
    struct LayerPkt_s Pkt;
    struct LayerExtSeq_s ExtSeq;
    struct LayerExtSeqDisp_s SeqDisp;
    struct LayerExtSeqDispColor_s SeqDispC;
    struct LayerExtCopyright_s Copyright;
    struct LayerExtPicDisp_s PicDisp;
    BYTE RawData[1];
  }u;
} LAYER;

#define SIGNATURE 0xCD010002
typedef struct AuxData_s {
  DWORD Signature;
  PHW pHw;
  BYTE CurLayer;		//Current layer being read. One of LAYER_
  BYTE CurExtLayer;		//Current extension layer. Is this really necessary?
  DWORD MaxData;		//The size of the largest data layer.
  DWORD ZeroTime;		//Beginning of time
  DWORD CurTime;		//Current PC time (milliseconds)
  DWORD LastEventVid;		//Last video event (picture start, slice start,...) One of VIDEVENT_
  DWORD TotalTimeouts;		//Total number of timeouts seen in all layers.
  union {
    struct {
      DWORD Rsvd:1;
    }s;
    DWORD Flags;
  }u;
  struct DataLayer_s Layers[LAYER_MAX];
} AUX;

/*************************************************************************/
/** Static Data								**/
/*************************************************************************/
static AUX StaticAux;

//Debugging logs
static WORD AuxDataLog[256];		//For debugging, record last 256 bytes read. Low byte is Data, Hi byte is status.
static BYTE AuxDataLogIndex;
static BYTE AuxPicTypeLog[256];		//PicureTypes, entry recorded for every LAYER_PIC completed.
static BYTE AuxPicTypeLogIndex;
static BYTE AuxPicTypeChr[]= {"0IPB4567"};  //Translates PicType index to I,P,B frames.

static struct IdentEntry_s IdentsAuxData[]= {
  { ID_AUXCMD, 0 },
    { AUXCMD_INIT, "AuxInit" },
    { AUXCMD_FINI, "AuxFini" },
    { AUXCMD_RESET, "AuxReset" },
    { AUXCMD_TIMER, "AuxTimer" },
    { AUXCMD_GET, "Get" },
  { ID_AUXLAYER, 0 },
    { LAYER_NONE, "None" },
    { LAYER_SEQ, "Sequence" },
    { LAYER_GOP, "GOP" },
    { LAYER_PIC, "Picture" },
    { LAYER_PKT, "Packet" }, 
    { LAYER_EXT, "Extension" },
    { LAYER_EXTSEQ, "SequenceExt" },
    { LAYER_EXTSDISP, "SequenceDisplayExt" }, 
    { LAYER_EXTQTBL, "QuantizerMatrix" }, 
    { LAYER_EXTCOPYRT, "Copyright" },
    { LAYER_EXTPDISP, "PictureDisplayExt" },
    { LAYER_EXTPIC, "PictureExt" },
  { 0,0 }
};

/*************************************************************************/
/** Handle conversions.							**/
/*************************************************************************/
//Convert an AuxData handle to an AUX pointer.
static AUX *AuxPtr(DWORD hAux) {
  if(hAux && ((AUX*)hAux)->Signature==SIGNATURE)
    return((AUX*)hAux);
  return(0);
}

//Convert a Layer ID to a pointer to Layer struct.
static LAYER *AuxLayerPtr(AUX *pAux, BYTE LayerID) {
  DWORD n1;
  for(n1=0;n1<LAYER_MAX;n1++) {
    if(pAux->Layers[n1].ID==LayerID)
      return(&pAux->Layers[n1]);
  }
  return(0);
}

/*************************************************************************/
/**	Only exported entry point.					**/
/*************************************************************************/
static long AuxInit(PHW pHw);
static long AuxFini(AUX *pAux);
static long AuxReset(AUX *pAux);
static long AuxTimer(AUX *pAux);
static long AuxGet(AUX *pAux, DWORD ItemID, DWORD *pDst);
static long AuxBadCmd(AUX *pAux, DWORD Cmd);

long AuxDataCmd(PHW pHw, DWORD Cmd, DWORD Arg1, DWORD Arg2) {
  long Err= 0;
  AUX *pAux;
  if(LOWORD(Cmd)==AUXCMD_INIT) {
    Err= AuxInit(pHw);
  } else if(!(pAux= AuxPtr(pHw->hAuxData))) {
    Err= MARGI_ERR_BADHANDLE;
  } else {
    pAux->CurTime= MargiGetTime()-pAux->ZeroTime;
    if(LOWORD(Cmd)==AUXCMD_TIMER) {
      Err= AuxTimer(pAux);
    } else {
      DbgMsg(DBG_AUXCMD,"AuxCmd %s.\r\n",Ident(ID_AUXCMD,LOWORD(Cmd)));
      switch(LOWORD(Cmd)) {
	case AUXCMD_FINI:	Err= AuxFini(pAux); break;
	case AUXCMD_RESET:	Err= AuxReset(pAux); break;
	case AUXCMD_GET:	Err= AuxGet(pAux,Arg1,(DWORD*)Arg2); break;
	default:		Err= AuxBadCmd(pAux,Cmd);
      }
    }
  }
  return(Err);
}

static long AuxBadCmd(AUX *pAux, DWORD Cmd) {
  DbgMsg(DBG_AUXERR,"Aux: Bad command %08X.\r\n",Cmd);
  return(MARGI_ERR_BADCMD);
}

/*************************************************************************/
/** Create/destroy/reset						**/
/*************************************************************************/
//Create a new auxillary data fifo object.
static long AuxInit(PHW pHw) {
  long Err= 0;
  AUX *pAux;
  DWORD n1;
#ifdef _WIN98_ONLY_
  IdentAddClass(IdentsAuxData);
#endif  //_WIN98_ONLY_
  //Eventually replace this with dynamic memory allocation.
  pAux= &StaticAux;
  pHw->hAuxData= (DWORD)pAux;
  pAux->Signature= SIGNATURE;
  pAux->pHw= pHw;
  pAux->MaxData= 0;
  for(n1=0;n1<LAYER_MAX;n1++) {
    pAux->Layers[n1].ID= LayersInUse[n1].LayerID;
    pAux->Layers[n1].Len= LayersInUse[n1].Len;
    pAux->Layers[n1].Timeout= LayersInUse[n1].Timeout;
    if(LayersInUse[n1].Len>pAux->MaxData)
      pAux->MaxData= LayersInUse[n1].Len;
  }
  AuxReset(pAux);
  return(Err);
}

//Destroy an auxillary data fifo object.
static long AuxFini(AUX *pAux) {
  long Err= 0;
  pAux->pHw->hAuxData= 0;
  pAux->Signature= 0;
  //Eventually free dynamically allocated pAux.
  return(Err);
}

static long AuxReset(AUX *pAux) {
  long Err= 0;
  DWORD n1;
  dvHwMainWriteLsiReg(pAux->pHw,64,1);			//Reset Aux data FIFO
  pAux->CurLayer= LAYER_NONE;
  pAux->CurExtLayer= 0;
  pAux->ZeroTime= MargiGetTime();
  pAux->CurTime= 0;
  pAux->LastEventVid= 0;
  pAux->TotalTimeouts= 0;
  for(n1=0;n1<LAYER_MAX;n1++) {;			//Clear data for each layer.
    pAux->Layers[n1].CurByte= 0;
    pAux->Layers[n1].TimeoutCnt= 0;
    pAux->Layers[n1].PcTime= 0;
  }
  return(Err);
}

/*************************************************************************/
/**	Check for data ready in FIFO, read it if present.		**/
/**	This needs to be called every interrupt, every time the status	**/
/**	registers are read. VERY IMPORTANT!				**/
/*************************************************************************/
static BYTE AuxReadData(AUX *pAux, BYTE *pLayer, BYTE *pData);
static long AuxChkEvents(AUX *pAux);
static long AuxLayerStart(AUX *pAux, LAYER *pLayer);
static long AuxLayerChkOverrun(AUX *pAux, LAYER *pLayer);
static long AuxLayerComplete(AUX *pAux, LAYER *pLayer);

static long AuxTimer(AUX *pAux) {
  long Err= 0;
  BYTE Layer,Data;
  LAYER *pLayer;
  AuxChkEvents(pAux);                       //Update event flags. Picture start, slice start, etc.
  while(AuxReadData(pAux,&Layer,&Data)) {     //Empty FIFO each time.
    if(!(pLayer= AuxLayerPtr(pAux,Layer))) {    //We do not record all layers.
      pAux->CurLayer= Layer;                      //If not recording this layer, simply record its ID.
    } else {
      if(Layer!=pAux->CurLayer)                     //Started a new layer?
        AuxLayerStart(pAux,pLayer);                   //Called when a new layer is started.
      if(!(Err= AuxLayerChkOverrun(pAux,pLayer))) { //Make sure there is room for new byte.
        pLayer->u.RawData[pLayer->CurByte++]= Data;     //Record data.
      if(pLayer->Len && pLayer->CurByte>=pLayer->Len)//Completed the layer?
        AuxLayerComplete(pAux,pLayer);                //Called when a layer has been completed.
      }
    }
  }
  return(Err);
}

//Low-level function to retrieve a single byte from the FIFO.
static BYTE AuxReadData(AUX *pAux, BYTE *pLayer, BYTE *pData) {
  BYTE DataReady,Status,Layer,Data;
  Status= dvHwMainReadLsiReg(pAux->pHw,64);		//Check FIFO status, data layer ID.
  Layer= (Status>>2) & 7;				//Extract data layer
  switch(Status & 3) {					//FIFO data status
    case 0:		DataReady= 0; break;		//FIFO empty
    case 1:		DataReady= 1; break;		//Data ready
    case 2:		DataReady= 1; break;		//FIFO full
    case 3:		DataReady= 0;			//FIFO overrun
			DbgMsg(DBG_AUXERR,"AuxData: FIFO overrun.\r\n");
			AuxReset(pAux); 
			break;
  }
  if(DataReady) {
    Data= dvHwMainReadLsiReg(pAux->pHw,67);		//Read Aux data FIFO.
    AuxDataLog[AuxDataLogIndex++]= ((WORD)Status<<8)|Data;	//Record everything for debugging.
    if(Layer==LAYER_EXT && !pAux->CurExtLayer) {	//Is this the first byte of an extension layer?
      pAux->CurExtLayer= Data+LAYER_EXT;		//Set current extension layer ID (first byte of data)
      DataReady= AuxReadData(pAux,pLayer,pData);	//Then get the next byte, or return no DataReady
    } else {
      if(Layer==LAYER_EXT)
	Layer= pAux->CurExtLayer;
      if(pLayer)
	*pLayer= Layer;
      if(pData)
	*pData= Data;
    }
  }
  return(DataReady);
}

//Update the various event flags.
static long AuxChkEvents(AUX *pAux) {
  if(pAux->pHw->bStatusFlag[1] & 2) {			//Check for Picture Start Code Detect
    pAux->LastEventVid|= EVENTVID_PICTSTART;
    pAux->LastEventVid&= ~EVENTVID_SLICESTART;
  }
  if(pAux->pHw->bStatusFlag[0] & 4)			//Check for First Slice Start Code Detect
    pAux->LastEventVid|= EVENTVID_SLICESTART;
  return(0);
}

//Start recording a new layer.
static long AuxLayerStart(AUX *pAux, LAYER *pNewLayer) {
  DWORD Elapsed;
  pAux->CurLayer= pNewLayer->ID;			//Set the current layer ID.
  pNewLayer->CurByte= 0;				//Reset the bytes read count.
  if(pNewLayer->PcTime) {				//Do we have a previous time?
    Elapsed= pAux->CurTime-pNewLayer->PcTime;		//Determine time elapsed since this layer was last recorded.
    if(pNewLayer->Timeout && Elapsed>=pNewLayer->Timeout) {	//Check for timeouts (missed pictures)
      DbgMsg(DBG_AUXMISS,"AuxData: Timeout %s %ums.\r\n",Ident(ID_AUXLAYER,pNewLayer->ID),Elapsed);
      pNewLayer->TimeoutCnt++;
      pAux->TotalTimeouts++;
    }
  }
  pNewLayer->PcTime= pAux->CurTime;			//Record time at which layer was started.
  return(0);
}

//Make sure we don't overrun our storage buffers.
static long AuxLayerChkOverrun(AUX *pAux, LAYER *pLayer) {
  long Err;
  if(pLayer->CurByte>=pAux->MaxData) {
    DbgMsg(DBG_AUXERR,"AuxData: Layer %s MaxData overrun at %u bytes.\r\n",Ident(ID_AUXLAYER,pLayer->ID),pLayer->CurByte);
    Err= MARGI_ERR_TOOLONG;
  } else if(pLayer->Len && pLayer->CurByte>=pLayer->Len) {
    DbgMsg(DBG_AUXERR,"AuxData: Layer %s overrun at %u bytes.\r\n",Ident(ID_AUXLAYER,pLayer->ID),pLayer->CurByte);
    Err= MARGI_ERR_TOOLONG;
  } else {
    Err= 0;
  }
  return(Err);
}

//Called when a layer has been completed. Mostly for debugging.
static long AuxLayerComplete(AUX *pAux, LAYER *pLayer) {
  DbgMsg(DBG_AUXLAYER|DBG_HIFREQ,"AuxData: %s.\r\n",Ident(ID_AUXLAYER,pLayer->ID));
  if(pLayer->ID==LAYER_PIC) {				//For picture layers, log the picture type.
    AuxPicTypeLog[AuxPicTypeLogIndex++]= AuxPicTypeChr[pLayer->u.Pic.PicType & 7];
    if(AuxPicTypeLogIndex>=48) {
      AuxPicTypeLog[AuxPicTypeLogIndex]= 0;
      DbgMsg(DBG_AUXPIC,"AuxData: PicType %s.\r\n",AuxPicTypeLog);
      AuxPicTypeLogIndex= 0;
    }
  }
  pAux->CurLayer= 0;
  pAux->CurExtLayer= 0;
  return(0);
}

/*************************************************************************/
/** Retrieve data							**/
/*************************************************************************/
static long AuxGetPicType(AUX *pAux, DWORD *pDst);

static long AuxGet(AUX *pAux, DWORD ItemID, DWORD *pDst) {
  long Err= 0;
  DWORD Data= 0;
  switch(ItemID) {
    case AUXGET_PICTYPE:	Err= AuxGetPicType(pAux,&Data); break;
    case AUXGET_EVENTVID:	Data= pAux->LastEventVid; break;
    default:			Err= MARGI_ERR_BADITEM;
  }
  if(!Err) {
    if(!pDst)
      Err= (long)Data;
    else
      *pDst= Data;
  }
  return(Err);
}

//Return the current picture type. 
static long AuxGetPicType(AUX *pAux, DWORD *pDst) {
  long Err;
  LAYER *pLayer;
  if(!(pLayer= AuxLayerPtr(pAux,LAYER_PIC))) {
    Err= MARGI_ERR_BADITEM;
  } else {
    *pDst= pLayer->u.Pic.PicType;
    Err= 0;
  }
  return(Err);
}

/**EOF: AUXDATA.C**/
