#include "system.h"
#include "DbgMsg.h"
#include "DbgFltr.h"
#include "Settings.h"

#include "Lsi64020.h"
#include "MidasApi.h"
#include "dvHwMain.h"
#include "Global.h"
#include "margierr.h"
#include "Idents.h"
#include "identcls.h"
#include "IniNames.h"
#include "dvHwOsd.h"

#define DEFAULT_OSD_SIZE 80000
#define MAX_REGIONS 16

#define OSD_TOP_OFFSET    50
#define OSD_LEFT_OFFSET  124

typedef struct OsdHdr_s {
  int   fInUse;     // Used instead of dynamic allcation
  int   fActive;    // This region is allocated in frame buffer memory
  WORD  wHiCol;     // Hi colour mode 2bits 0:2bps, 1:8bps, 2:4bps
  WORD  wStartRow;  // 1st line of region 9 bits
  WORD  wEndRow;    // Last line of region  9bits
  WORD  wStartCol;  // 1st pixel of region 9 bits
  WORD  wEndCol;    // Last pixel of region 9 bits
  WORD  wBmOnly;    // True if no palette in region
  WORD  wMix;       // Mix weight 4bits 
  WORD  wRsvd;
  DWORD dwFieldAddr[2];//Address in LSI memory of header for each field
  DWORD dwFieldSize[2];//Size in bytes of field bitmap in LSI memory
  DWORD dwBitmapStartAddr[2]; //Address in LSI memory of bitmap for each field
  DWORD dwBitmapEndAddr[2];   //Address in LSI memory of bitmap end for each field

//  DWORD dwLsiSize;  // Number of bytes required in LSI memory
  DWORD dwNxtAddr;  // Start address of next region in LSI
  WORD  Pal[16];    // Colour palette (supports upto 4bps)
  int   fShow;      // Display region if TRUE
  struct OsdHdr_s* pPrvHdr;   // Link to prev Region Hdr 
  struct OsdHdr_s* pNxtHdr;   // Link to next Region Hdr 
} OSDHDR,*POSDHDR;

typedef struct OsdD_s {
  DWORD dwAddr;       // Dram addr of register index
  BYTE  bPos;
  BYTE  bFlags;       // 0=Dram data, 1=Register, 2=Fill
  WORD  wBytes;
  BYTE  Data[360];
} OSDDAT, *POSDDAT;

#define OSD_QUEUE_SIZE 100
typedef struct OsdQ_s {
  int fInUse;
  int nHead;
  int nTail;
  int nEntries;
  OSDDAT Dat[OSD_QUEUE_SIZE];
} OSDQUEUE;


#define SIGNATURE 0xCD010006
typedef struct Osd_s {
  DWORD Signature;
  PHW   pHw;
  DWORD CurTime;
  DWORD TimeZero;
  DWORD dwTerminateHdr;         // Address of OSD Terminate Header
  OSDHDR Headers[MAX_REGIONS];  // Use array of headers instead of dynamic allocation
  POSDHDR pHead;                // Pointer to 1st header in linked list
  POSDHDR pTail;                // Pointer to last header in linked list
  OSDQUEUE DataQueue;
} OSD,*POSD;

OSD StaticOsd;

static OSD *OsdPtr(PVOID hOsd);
static long OsdCmdInvalid(POSD pOsd, DWORD Cmd);
static long OsdInit (PHW pHw);
static long OsdFini (POSD pOsd);
static long OsdTimer(POSD pOsd);
static long OsdReset(POSD pOsd);
static long OsdSetBuffSize(POSD pOsd,DWORD dwBytes);
static long OsdOpenRegion (POSD pOsd,HOSDREGION* hReg,POSDREGION pRegDesc);
static long OsdCloseRegion(POSD pOsd,HOSDREGION hReg);
static long OsdFillRegion (POSD pOsd,HOSDREGION hReg,POSDSTRIP pStrip);
static long OsdFillRect   (POSD pOsd,HOSDREGION hReg,POSDSTRIP pStrip);
static long OsdRenderStrip(POSD pOsd,HOSDREGION hReg,POSDSTRIP pStrip);
static long OsdRenderRect (POSD pOsd,HOSDREGION hReg,POSDSTRIP pRect);
static long OsdShowRegion (POSD pOsd,HOSDREGION hReg);
static long OsdHideRegion (POSD pOsd,HOSDREGION hReg);
static long OsdSetPalette (POSD pOsd,HOSDREGION hReg);
static DWORD OsdWrFieldHeader(POSD pOsd,int nField,DWORD dwAddr,POSDHDR pHdr);
static long OsdConfigAll(POSD pOsd);
static void OsdUpdateAll(POSD pOsd);
static long OsdQueueData(POSD pOsd,DWORD dwAddr,BYTE* pBuff,BYTE bPos,WORD wBytes,BYTE bFlags);
static DWORD OsdCalcWidth(POSDHDR pHdr);
static DWORD OsdCalcWidthPix(POSDHDR pHdr);
static int OsdGetField(POSDHDR pHdr,WORD wRow);

static OSD *OsdPtr(PVOID hOsd) {
  if(hOsd && *(DWORD*)hOsd==SIGNATURE)
    return((OSD*)hOsd);
  return(0);
}

long OsdCmd(PHW pHw, DWORD Cmd, DWORD Arg1, DWORD Arg2) {
  OSD *pOsd;
  long Err;

  if(LOWORD(Cmd)==OSDCMD_INIT) {
    Err= OsdInit(pHw);
  } else if(!(pOsd= OsdPtr(pHw->hOsd))) {
    Err= MARGI_ERR_BADHANDLE;
  } else {
    pOsd->CurTime= MargiGetTime()-pOsd->TimeZero;
    if(LOWORD(Cmd)==OSDCMD_TIMER) {
      Err= OsdTimer(pOsd);
    } else {
      switch(LOWORD(Cmd)) {
        case OSDCMD_FINI:        Err= OsdFini(pOsd);  break;
//        case OSDCMD_RESET:       Err= OsdReset(pOsd); break;
        case OSDCMD_SETBUFFSIZE: Err= OsdSetBuffSize(pOsd,Arg1); break;
        case OSDCMD_OPEN_REGION: Err= OsdOpenRegion (pOsd,(HOSDREGION*)Arg1,(POSDREGION)Arg2); break;
        case OSDCMD_CLOSE_REGION:Err= OsdCloseRegion(pOsd,(HOSDREGION)Arg1); break; 
        case OSDCMD_FILL_REGION: Err= OsdFillRegion (pOsd,(HOSDREGION)Arg1,(POSDSTRIP)Arg2); break;
        case OSDCMD_FILL_RECT:   Err= OsdFillRect   (pOsd,(HOSDREGION)Arg1,(POSDSTRIP)Arg2); break;
        case OSDCMD_RENDER_STRIP:Err= OsdRenderStrip(pOsd,(HOSDREGION)Arg1,(POSDSTRIP)Arg2); break;
        case OSDCMD_RENDER_RECT: Err= OsdRenderRect (pOsd,(HOSDREGION)Arg1,(POSDSTRIP)Arg2); break;
        case OSDCMD_SHOW_REGION: Err= OsdShowRegion (pOsd,(HOSDREGION)Arg1); break;
        case OSDCMD_HIDE_REGION: Err= OsdHideRegion (pOsd,(HOSDREGION)Arg1); break;
        case OSDCMD_SETPALETTE:  Err= OsdSetPalette (pOsd,(HOSDREGION)Arg1); break;

        default:		Err= OsdCmdInvalid(pOsd,Cmd);
      }
    }
  }
  return(Err);
}

static long OsdCmdInvalid(POSD pOsd,DWORD Cmd) {
  DbgMsg(DBG_ERR,"Osd: Bad command %u (%X).\r\n",LOWORD(Cmd),Cmd);
  return(MARGI_ERR_BADCMD);
}

static long OsdInit(PHW pHw) {
  long Err= 0;
  OSD *pOsd;
  pOsd= &StaticOsd;					//Eventually replace with dynamic memory allocation.
  if(!pOsd) {
    DbgMsg(DBG_ERR,"Osd: Unable to allocate new OSD object!\r\n");
    Err= MARGI_ERR_NOMEM;
  } else {
    pOsd->Signature= SIGNATURE;
    pOsd->pHw= pHw;
    pHw->hOsd= (PVOID)pOsd;
    pHw->fOsdDisable=0;
//    pHw->fOsdMinMode=0;
    dvHwOsdBuffReserve(pHw,DEFAULT_OSD_SIZE);

    pOsd->TimeZero= MargiGetTime();
//    pHw->dwOsdStartAddr=0;
//   pHw->dwOsdEndAddr=0x40000;
  }
  return(Err);
}

static long OsdFini(POSD pOsd) {
  pOsd->Signature= 0;
  pOsd->pHw->hOsd= 0;
  dvHwOsdBuffReserve(pOsd->pHw,0);
//  pOsd->pHw->dwOsdStartAddr=0;
//  pOsd->pHw->dwOsdEndAddr=0;
  //Free dynamically allocated memory.
  return(0);
}

static long OsdSetBuffSize(POSD pOsd,DWORD dwBytes){
  return dvHwOsdBuffReserve(pOsd->pHw,dwBytes);
}

static long OsdOpenRegion(POSD pOsd,HOSDREGION* hReg,POSDREGION pRegDesc){
  long Err=MARGI_ERR_TOOMANY;
  int  bits[4]={2,8,4,0};
  POSDHDR pHdr;
  int n;
  *hReg=0;
  for(n=0;n<MAX_REGIONS;n++) {
    if(!pOsd->Headers[n].fInUse){
      pHdr=&pOsd->Headers[n];
      break;
    }
  }
  if(n<MAX_REGIONS) {
    pHdr->fInUse=1;
    pHdr->wHiCol=pRegDesc->wClrMode&3;      // Hi colour mode 2bits 0:2bps, 1:8bps, 2:4bps

    //force width to be integral # of 64bit words
    if(pRegDesc->dx*bits[pHdr->wHiCol]&0x3f) {
      DWORD dwBits=pRegDesc->dx*bits[pHdr->wHiCol]+63;
      dwBits&=~0x3f;
      pRegDesc->dx=(WORD)(dwBits/bits[pHdr->wHiCol]);
    }

    // force height to be multiple of 2 lines (same # lines in each field)
    pRegDesc->dy=(pRegDesc->dy+1)&~1;

    pHdr->wStartCol=pRegDesc->x+OSD_LEFT_OFFSET;
    pHdr->wStartRow=pRegDesc->y+OSD_TOP_OFFSET;
    pHdr->wEndCol=pHdr->wStartCol+pRegDesc->dx-1;
    pHdr->wEndRow=pHdr->wStartRow+pRegDesc->dy-1;


    pHdr->wBmOnly=0;                        // True if no palette in region
    pHdr->wMix=pRegDesc->wMixWeight&0x0f;   // Mix weight 4bits
    for(n=0;n<sizeof(pHdr->Pal)/sizeof(pHdr->Pal[0]);n++)
      pHdr->Pal[n]=pRegDesc->Palette[n];    // Copy palette

    pHdr->fShow=0;
    if(pOsd->pTail) {
      pOsd->pTail->pNxtHdr=pHdr;
      pHdr->pNxtHdr=0;
      pHdr->pPrvHdr=pOsd->pTail;
      pOsd->pTail=pHdr;
    } else {
      pOsd->pHead=pHdr;
      pOsd->pTail=pHdr;
      pHdr->pNxtHdr=0;
      pHdr->pPrvHdr=0;
    }
    Err=OsdConfigAll(pOsd);
    if(Err!=0)
      OsdCloseRegion(pOsd,(HOSDREGION)pHdr);
    else
      *hReg=pHdr;
  }  
  return Err;
}

static long OsdCloseRegion(POSD pOsd,HOSDREGION hReg){
  long Err=0;
  POSDHDR p;
  POSDHDR pHdr=(POSDHDR)hReg;

  //Remove hdr link
  for(p=pOsd->pHead;p;p=p->pNxtHdr) {
    if(p==pHdr) {
      p->fInUse=0;
      if(pOsd->pHead==p)
        pOsd->pHead=p->pNxtHdr;
      if(pOsd->pTail==p)
        pOsd->pTail=p->pPrvHdr;
      if(p->pPrvHdr)
        p->pPrvHdr->pNxtHdr=p->pNxtHdr;
      if(p->pNxtHdr)
        p->pNxtHdr->pPrvHdr=p->pPrvHdr;
      break;
    }
  }
  Err=OsdConfigAll(pOsd);
  return Err;
}

static long OsdRenderStrip(POSD pOsd,HOSDREGION hReg, POSDSTRIP pStrip){
  int  bits[4]={2,8,4,0};
  long Err=0;
  POSDHDR pHdr=(POSDHDR)hReg;
  int nField;
  int nPos,nBytes;
  DWORD dwPitch,dwAddr;
//  int nX;
//  int nY;
//  BYTE* pStrip;     // upto 8 bytes
//  BYTE bSize;       // number of bytes

  //compute Dram address from region Coordinates
  nPos=pStrip->x&0x07;
  pStrip->x&=~0x07;
  nField=OsdGetField(pHdr,(WORD)pStrip->y);
  dwPitch=OsdCalcWidthPix(pHdr);
  dwAddr= (pStrip->y/2)*dwPitch;
  dwAddr+=((pStrip->x*bits[pHdr->wHiCol&3])+63)/64;
  dwAddr+=pHdr->dwBitmapStartAddr[nField];
  nBytes=pStrip->wSize/2;
  while(nBytes) {
    if(dwAddr>=pHdr->dwBitmapEndAddr[nField])
      break;
    if(nBytes<8 || nPos!=0) {
      int n=(nBytes<8-nPos)?nBytes:8-nPos;
      OsdQueueData(pOsd,dwAddr,pStrip->pStrip,(BYTE)nPos,(BYTE)n,OSD_FLAG_COPY);
      nPos=0;
      pStrip->pStrip+=n;
      nBytes-=n;
    } else if(nBytes>=16) {
      OsdQueueData(pOsd,dwAddr,pStrip->pStrip,0,16,OSD_FLAG_COPY);
      pStrip->pStrip+=16;
      nBytes-=16;
      dwAddr++;
    } else {
      OsdQueueData(pOsd,dwAddr,pStrip->pStrip,0,8,OSD_FLAG_COPY);
      pStrip->pStrip+=8;
      nBytes-=8;
    }
    dwAddr++;
  }
  return Err;
}

static long OsdRenderRect(POSD pOsd,HOSDREGION hReg,POSDSTRIP pRect) {
  int  bits[4]={2,8,4,0};
  POSDHDR pHdr=(POSDHDR)hReg;
  DWORD dwAddr;
  DWORD dwPitchPix;
  DWORD dwPitch;
  WORD  wRow;
  WORD  wWidth;
  BYTE* pData;
  int nField;
  pRect->x&=~0x07;
  dwPitchPix=OsdCalcWidthPix(pHdr);
  dwPitch = OsdCalcWidth(pHdr);
  wWidth=pRect->dx/2;
  pData=pRect->pStrip + pRect->y*(dwPitch/2) + pRect->x/2;
//  pData = pRect->pStrip;
  for(wRow=pRect->y;wRow<pRect->y+pRect->dy;wRow++) {
    nField=OsdGetField(pHdr,wRow);
    dwAddr= (wRow/2)*dwPitchPix;
    dwAddr+=((pRect->x*bits[pHdr->wHiCol&3])+63)/64;
    dwAddr+=pHdr->dwBitmapStartAddr[nField];
    OsdQueueData(pOsd,dwAddr,pData,0,wWidth,OSD_FLAG_COPY);
    pData+=(dwPitch/2);
//    pData += wWidth;
  }
  return 0;
}

static long OsdFillRect(POSD pOsd,HOSDREGION hReg,POSDSTRIP pRect) {
  int  bits[4]={2,8,4,0};
  POSDHDR pHdr=(POSDHDR)hReg;
  DWORD dwAddr;
  DWORD dwPitchPix;
  DWORD dwPitch;
  WORD  wRow;
  WORD  wWidth;
  BYTE* pData;
  int nField;
  pRect->x&=~0x07;
  dwPitchPix=OsdCalcWidthPix(pHdr);
  dwPitch = OsdCalcWidth(pHdr);
  wWidth=pRect->dx/2;
  pData=pRect->pStrip;
  for(wRow=pRect->y;wRow<pRect->y+pRect->dy;wRow++) {
    nField=OsdGetField(pHdr,wRow);
    dwAddr= (wRow/2)*dwPitchPix;
    dwAddr+=((pRect->x*bits[pHdr->wHiCol&3])+63)/64;
    dwAddr+=pHdr->dwBitmapStartAddr[nField];
    OsdQueueData(pOsd,dwAddr,pData,0,wWidth,OSD_FLAG_SET);
  }
  return 0;
}

static long OsdFillRegion(POSD pOsd,HOSDREGION hReg,POSDSTRIP pStrip) {
  POSDHDR pHdr=(POSDHDR)hReg;
  DWORD dwPitch;
  DWORD dwAddr[2];
  WORD  wRow;
  int nField;
  dwAddr[0]=pHdr->dwBitmapStartAddr[0];
  dwAddr[1]=pHdr->dwBitmapStartAddr[1];
  dwPitch=OsdCalcWidthPix(pHdr);
  for(wRow=0;wRow<=pHdr->wEndRow-pHdr->wStartRow;wRow++) {
    nField=OsdGetField(pHdr,wRow);
    OsdQueueData(pOsd,dwAddr[nField],pStrip->pStrip,0,(WORD)(dwPitch*8),OSD_FLAG_SET);
    dwAddr[nField]+=dwPitch;
  }
  return 0;
}

static long OsdShowRegion(POSD pOsd,HOSDREGION hReg){
  long Err=0;
  POSDHDR pHdr=(POSDHDR)hReg;
  if(pHdr)
    pHdr->fShow=1;
  else
    for(pHdr=pOsd->pHead;pHdr;pHdr=pHdr->pNxtHdr)
      pHdr->fShow=1;
  OsdUpdateAll(pOsd);
  return Err;
}

static long OsdHideRegion(POSD pOsd,HOSDREGION hReg){
  long Err=0;
  POSDHDR pHdr=(POSDHDR)hReg;
  if(pHdr)
    pHdr->fShow=0;
  else
    for(pHdr=pOsd->pHead;pHdr;pHdr=pHdr->pNxtHdr)
      pHdr->fShow=0;
  OsdUpdateAll(pOsd);
  return Err;
}

static long OsdSetPalette(POSD pOsd,HOSDREGION hReg){
  long Err=0;
  return Err;
}

#ifdef _DUMP
int gfDumpSwitch=0;
BYTE DramBuffer[16384];
#endif

static long OsdTimer(POSD pOsd){
  long Err=0;
  DWORD dwTotalBytes=0;
  int  n;
  POSDDAT pDat;
  PHW pHw=pOsd->pHw;
  BYTE buff[8];
  if(pHw) {
    if(pOsd->DataQueue.fInUse==0) {
      while(pOsd->DataQueue.nEntries) {
        if(dwTotalBytes>pOsd->pHw->dwOsdMaxBytesPerTick)
          break;
        pDat=&pOsd->DataQueue.Dat[pOsd->DataQueue.nHead];
        if(pDat->bFlags&OSD_FLAG_WR_REG) {          // if write to OSD reg
          for(n=0;n<pDat->wBytes;n++) {
            BYTE bVal;
            // if reg 265 then only modify OSD part
            if(pDat->dwAddr==265) {
              bVal=dvHwMainReadLsiReg(pHw,(WORD)pDat->dwAddr);
              pDat->Data[n]=(pDat->Data[n]&0x0f)|(bVal&0xf0);
            }
            dvHwMainWriteLsiReg(pHw,(WORD)pDat->dwAddr++,pDat->Data[n]);
          }
        } else if(pDat->bFlags&OSD_FLAG_SET) {      // if set with repeated 8 bytes
          dvHwMainFillDram(pHw,pDat->dwAddr*8,pDat->Data,pDat->wBytes);
        } else {                                    // if copy wBytes of data
          BYTE* ptr=pDat->Data;
          n=pDat->wBytes;
          if(pDat->wBytes<8) {
            dvHwMainReadDram(pHw,pDat->dwAddr*8,buff,8);
            for(n=pDat->bPos;n<pDat->bPos+pDat->wBytes;n++) {
              buff[n]=pDat->Data[n];
            }
            ptr=buff;
            n=8;
          }
          dvHwMainWriteDram(pHw,pDat->dwAddr*8,ptr,n);
#ifdef _DUMP
          if(gfDumpSwitch) {
            DWORD dwAddr=OsdGetBaseAddr(pOsd)-OsdGetBuffSize(pOsd);
            dvHwMainReadDram(pHw,dwAddr*8,DramBuffer,sizeof(DramBuffer));
            dvHwMainWriteDram(pHw,dwAddr*8,DramBuffer,sizeof(DramBuffer));
          }
#endif
        }
        pOsd->DataQueue.nEntries--;
        if(++pOsd->DataQueue.nHead>=OSD_QUEUE_SIZE)
          pOsd->DataQueue.nHead=0;
        dwTotalBytes+=pDat->wBytes;
      }
    }
  }
  return Err;
}

static long OsdQueueData(POSD pOsd,DWORD dwAddr,BYTE* pBuff,BYTE bPos,WORD wBytes,BYTE bFlags) {
  long Err=MARGI_ERR_TOOMANY;
  POSDDAT pDat;
  int n,nBytes;
  if(!pOsd->pHw->fOsdDisable) {
    while(pOsd->DataQueue.nEntries>=OSD_QUEUE_SIZE) {
      MargiSleep(10);
    }
    if(pOsd->DataQueue.nEntries<OSD_QUEUE_SIZE) {
      pOsd->DataQueue.fInUse=1;
      pDat=&pOsd->DataQueue.Dat[pOsd->DataQueue.nTail];
      pDat->dwAddr=dwAddr;
      pDat->bPos=bPos;
      pDat->wBytes=wBytes;
      pDat->bFlags=bFlags;
      if(bFlags&OSD_FLAG_SET)
        nBytes=8;
      else
        nBytes=(int)wBytes;
      for(n=0;n<nBytes;n++)
        pDat->Data[n+bPos]=pBuff[n];
      pOsd->DataQueue.nEntries++;
      if(++pOsd->DataQueue.nTail>=OSD_QUEUE_SIZE)
        pOsd->DataQueue.nTail=0;
      Err=0;
      pOsd->DataQueue.fInUse=0;
    }
  }
  return Err;
}

static DWORD OsdWrFieldHeader(POSD pOsd,int nField,DWORD dwAddr,POSDHDR pHdr) {
  int  bits[4]={2,8,4,0};
  PHW pHw=pOsd->pHw;
  BYTE buff[8];
  WORD wStartRow;
  WORD wEndRow;
  DWORD dwBytes;

  if(nField) {
    wStartRow=(pHdr->wStartRow&1)?pHdr->wStartRow/2    :(pHdr->wStartRow+1)/2;
    wEndRow=  (pHdr->wEndRow&1)  ?pHdr->wEndRow/2      :(pHdr->wEndRow+1)/2;
  } else {
    wStartRow=(pHdr->wStartRow&1)?(pHdr->wStartRow-1)/2:pHdr->wStartRow/2;
    wEndRow=  (pHdr->wEndRow&1)  ?(pHdr->wEndRow-1)/2  :pHdr->wEndRow/2;
  }
  //Calculate # bytes in bitmap
  dwBytes= pHdr->wEndCol-pHdr->wStartCol+1;
  dwBytes*=wEndRow-wStartRow+1;
  dwBytes*=bits[pHdr->wHiCol&0x03];
  dwBytes/=8;
  pHdr->dwFieldSize[nField]=dwBytes;

  buff[0]= (BYTE)((pHdr->wHiCol&0x02)<<6);
  buff[0]|=(BYTE)((pHdr->dwNxtAddr&0x3f)<<1);
  buff[0]|=(BYTE)((wStartRow&0x100)>>8);

  buff[1]=(BYTE)wStartRow;

  buff[2]= (BYTE)((pHdr->wMix&0x0f)<<4);
  buff[2]|=(BYTE)((pHdr->wHiCol&0x01)<<3);
  buff[2]|=(BYTE)((pHdr->wBmOnly&0x01)<<2);
  buff[2]|=(BYTE)((pHdr->dwNxtAddr&0x40000)>>(18-1));
  buff[2]|=(BYTE)((wEndRow&0x100)>>8);

  buff[3]= (BYTE)wEndRow;

  buff[4]= (BYTE)((pHdr->dwNxtAddr&0x0fc0)>>4);
  buff[4]|=(BYTE)((pHdr->wStartCol&0x0300)>>8);

  buff[5]= (BYTE)pHdr->wStartCol;

  buff[6]= (BYTE)((pHdr->dwNxtAddr&0x3f000)>>10);
  buff[6]|=(BYTE)((pHdr->wEndCol&0x0300)>>8);

  buff[7]= (BYTE)pHdr->wEndCol;

  pHdr->dwFieldAddr[nField]=dwAddr; 
  OsdQueueData(pOsd,dwAddr,buff,0,8,OSD_FLAG_COPY);
  return 1;  // # bytes in header plus bytes in bitmap
}

static DWORD OsdWrPalette(POSD pOsd,DWORD dwAddr,POSDHDR pHdr) {
  int cols[4]={4,256,16,0};
  WORD wBytes=cols[pHdr->wHiCol&0x03]*2; //#colours*2bytes
  BYTE* ptr=(BYTE*)pHdr->Pal;
  WORD w=wBytes;
  PHW pHw=pOsd->pHw;
  while(w) {
    OsdQueueData(pOsd,dwAddr++,ptr,0,8,OSD_FLAG_COPY);
    ptr+=8;
    w-=8;
  }
  return wBytes/8;
}

static DWORD OsdCalcMemReq(POSDHDR pHdr) {
  int  bits[4]={2,8,4,0};
  int  cols[4]={4,256,16,0};
  DWORD dw,dwBytes;
  DWORD dwFldHgt;

  //calc field bitmap size in bits
  dwFldHgt=(pHdr->wEndRow-pHdr->wStartRow+1+1)/2;
  dw= dwFldHgt*(pHdr->wEndCol-pHdr->wStartCol+1);
  dw*=bits[pHdr->wHiCol&0x03];
  //Calc field bitmap size in bytes
  dwBytes=dw/8;

  // field header is 8 bytes
  dwBytes+=8;

  //calc palette size
  dwBytes+=cols[pHdr->wHiCol&0x03]*2; //#colours*2bytes

  // each field must start on a 64 byte boundary
  // so round up to nearest 64 bytes
  dwBytes+=63;
  dwBytes&=~0x3f;

  // return entire region size in 64 bit words
  return dwBytes*2/8;
}

static DWORD OsdCalcWidthPix(POSDHDR pHdr) {
  DWORD dwWidth;
  int  bits[4]={2,8,4,0};
  dwWidth=(((pHdr->wEndCol-pHdr->wStartCol+1)*bits[pHdr->wHiCol&3])+63)/64;
  return dwWidth;
}

static DWORD OsdCalcWidth(POSDHDR pHdr) {
  DWORD dwWidth;
  dwWidth=(pHdr->wEndCol-pHdr->wStartCol+1);
  return dwWidth;
}
    
static int OsdGetField(POSDHDR pHdr,WORD wRow) {
  int fOddField1st,nField;
  fOddField1st=(pHdr->wStartRow&1)==0;
  nField=(wRow&1)^fOddField1st;
  return nField;
}

static POSDHDR OsdFindPrvShow(POSDHDR pHdr) {
  for(pHdr=pHdr->pPrvHdr;pHdr;pHdr=pHdr->pPrvHdr) {
    if(pHdr->fShow)
      break;
  }
  return pHdr;
}

static POSDHDR OsdFindNxtShow(POSDHDR pHdr) {
  for(pHdr=pHdr->pNxtHdr;pHdr;pHdr=pHdr->pNxtHdr) {
    if(pHdr->fShow)
      break;
  }
  return pHdr;
}

static DWORD OsdGetBaseAddr(POSD pOsd) {
//  if(pOsd->pHw->fOsdMinMode)
    return pOsd->pHw->dwOsdEndAddr;
//  else
//    return 0;
}

static DWORD OsdGetMemSize(POSD pOsd) {
//  if(pOsd->pHw->fOsdMinMode)
    return dvHwMainGetDramSize(pOsd->pHw);
    return pOsd->pHw->dwOsdEndAddr-pOsd->pHw->dwOsdStartAddr;
//  else
//    return dvHwMainGetDramSize(pOsd->pHw);
}

static void OsdUpdateAll(POSD pOsd) {
  long Err=0;
  PHW pHw=pOsd->pHw;
  POSDHDR pHdr,pPrvHdr,pNxtHdr;
  DWORD dwEvenStartAddr;
  DWORD dwOddStartAddr;
  DWORD dw;
  int nShow=0;
  dwEvenStartAddr=pOsd->dwTerminateHdr;  //in case nothing to show
  dwOddStartAddr= pOsd->dwTerminateHdr;  //point at terminate header
  for(pHdr=pOsd->pHead;pHdr;pHdr=pHdr->pNxtHdr) {
    if(pHdr->fShow) {
      nShow++;
      pPrvHdr=OsdFindPrvShow(pHdr);
      if(pPrvHdr==0) {
        dwEvenStartAddr=pHdr->dwFieldAddr[0];
        dwOddStartAddr= pHdr->dwFieldAddr[1];
      } else {
        pPrvHdr->dwNxtAddr=pHdr->dwFieldAddr[0];
        OsdWrFieldHeader(pOsd,0,pPrvHdr->dwFieldAddr[0],pPrvHdr);
        pPrvHdr->dwNxtAddr=pHdr->dwFieldAddr[1];
        OsdWrFieldHeader(pOsd,1,pPrvHdr->dwFieldAddr[1],pPrvHdr);
      }
      pNxtHdr=OsdFindNxtShow(pHdr);
      if(pNxtHdr==0) {
        //Write last header with link to Terminate Header
        pHdr->dwNxtAddr=pOsd->dwTerminateHdr;
        OsdWrFieldHeader(pOsd,0,pHdr->dwFieldAddr[0],pHdr);
        OsdWrFieldHeader(pOsd,1,pHdr->dwFieldAddr[1],pHdr);
      }
    }
  }
  //Write OSD field pointers
  dw=  dwOddStartAddr/8;
  dw+=(dwEvenStartAddr/8)<<16;
  OsdQueueData(pOsd,270,(BYTE*)&dw,0,4,OSD_FLAG_WR_REG);
  //Enable linked OSD mode is regions to show
  dw=nShow?0x0a:0x00;
  OsdQueueData(pOsd,265,(BYTE*)&dw,0,1,OSD_FLAG_WR_REG);
}

static long OsdConfigAll(POSD pOsd) {
  long Err=0;
  PHW pHw=pOsd->pHw;
  POSDHDR pHdr;
  BYTE  bTermHdr[8]={0x01,0xff,0x00,0x00,0x00,0x00,0x00,0x00};
  DWORD dwTopAddr= OsdGetBaseAddr(pOsd);  // OSD memory is allocated down from top of DRAM
  DWORD dwQuads=   OsdGetMemSize(pOsd);

#if 0
  {int n;
    for(n=0;n<sizeof(DramBuffer);n++)
      DramBuffer[n]=0xdd;
    dvHwMainWriteDram(pHw,0,DramBuffer,sizeof(DramBuffer));
  }
#endif
  // termination header is 8 bytes, use for both fields
  //dvHwMainWriteDram(pHw,dwAddr*8,bTermHdr,8);
  dwTopAddr-=8;
  OsdQueueData(pOsd,dwTopAddr,bTermHdr,0,8,OSD_FLAG_COPY);
  pOsd->dwTerminateHdr=dwTopAddr;
  dwQuads-=8;
  // write headers for all regions
  for(pHdr=pOsd->pHead;pHdr;pHdr=pHdr->pNxtHdr) {
    DWORD dwRegionQuads=OsdCalcMemReq(pHdr);
    pHdr->fActive=0; 
    if(dwRegionQuads<=dwQuads) {
      DWORD dwAddr;
      dwTopAddr-=dwRegionQuads;
      dwAddr=dwTopAddr;
      dwAddr+=OsdWrFieldHeader(pOsd,0,dwAddr,pHdr);
      dwAddr+=OsdWrPalette(pOsd,dwAddr,pHdr);
#if 0
  {int n;
    for(n=0;n<(int)pHdr->dwFieldSize[0];n++)
      DramBuffer[n]=0x88;
    dvHwMainWriteDram(pHw,dwAddr*8,DramBuffer,pHdr->dwFieldSize[0]);
  }
#endif
      pHdr->dwBitmapStartAddr[0]=dwAddr; 
      dwAddr+=pHdr->dwFieldSize[0]/8;
      pHdr->dwBitmapEndAddr[0]=dwAddr; 

      dwAddr+=7; dwAddr&=~7;
      dwAddr+=OsdWrFieldHeader(pOsd,1,dwAddr,pHdr);
      dwAddr+=OsdWrPalette(pOsd,dwAddr,pHdr);
      pHdr->dwBitmapStartAddr[1]=dwAddr; 
#if 0
  {int n;
    for(n=0;n<(int)pHdr->dwFieldSize[1];n++)
      DramBuffer[n]=0x99;
    dvHwMainWriteDram(pHw,dwAddr*8,DramBuffer,pHdr->dwFieldSize[1]);
  }
#endif
      dwAddr+=pHdr->dwFieldSize[1]/8;
      pHdr->dwBitmapEndAddr[1]=dwAddr; 
      dwAddr+=7; dwAddr&=~7;
      pHdr->fActive=1;
    }
    else {
      Err=MARGI_ERR_NOMEM;
    }
  }
  // re-write headers with links 
  // to show or hide regions
  OsdUpdateAll(pOsd);
  return Err;
}
